# 
# *****************************************************************
# *                                                               *
# *    Copyright Compaq Computer Corporation, 2001                *
# *                                                               *
# *   The software contained on this media  is  proprietary  to   *
# *   and  embodies  the  confidential  technology  of  Compaq    *
# *   Computer Corporation.  Possession, use,  duplication  or    *
# *   dissemination of the software and media is authorized only  *
# *   pursuant to a valid written license from Compaq Computer    *
# *   Corporation.                                                *
# *                                                               *
# *   RESTRICTED RIGHTS LEGEND   Use, duplication, or disclosure  *
# *   by the U.S. Government is subject to restrictions  as  set  *
# *   forth in Subparagraph (c)(1)(ii)  of  DFARS  252.227-7013,  *
# *   or  in  FAR 52.227-19, as applicable.                       *
# *                                                               *
# *****************************************************************
#
# HISTORY
# 
# @(#)$RCSfile: subset.tcl,v $ $Revision: 1.1.18.3 $ (DEC) $Date: 2000/06/26 18:56:30 $
# 

##
## MODULE:
##	subset.tcl
##
## ABSTRACT:
##	Routines to manage detailed info about subsets, in the
##	global Subsets() array.
##
## CONTAINS:
##	pcode_lower {ss_name}
##	product_load_details {product}
##	product_subsets_in_toc {product}
##	setld_already_done {product {err_ret_name {}}}
##	setld_load_details {product}
##	setld_load_outdone_file {product err_ret_name}
##	setld_s_start {product}
##	setld_s_wait {}
##	subset_copy_mand_from_cache {product}
##	subset_define {name}
##	subset_fast_flags {product context shell_init subset_list}
##	subset_hide_deprecated {product}
##	subset_init_fs_usage {}
##	subset_insert_toc_category {subset_list}
##	subset_list_all_dependencies {ss_list}
##	subset_list_all_enabled {ss_list}
##	subset_list_all_related {ss_list ss_field}
##	subset_list_auto_related {ss_list pick_flag}
##	subset_pick_handler {always_Subsets name_plus always_w}
##	subset_pick_list {pick_flag ss_list}
##	subset_update_all_fstotals {}
##	subset_update_fstotal {fs_list}
##
## NOTES:
##	The most important routine here is product_load_details{}.  Its
##	job is to run "setld -s" for a single product, and then record
##	all the info from the resulting /tmp/setld-s.outdone data file.
##	It loads the data into the global Subsets() array, which is
##	indexed by strings of the form "<subset_name>.<field>".  Here
##	are the various fields it creates for each subset:
##
##           Subset Field   Description
##           ------------   ---------------------------------------------
##           .mand          1 if subset is mandatory
##           .i18n.blksize  space needed in /usr/i18n
##           .root.blksize  space needed in root
##           .req_list      prerequisite subsets (DEPS)
##           .enables       subsets for which this subset is a prerequisite
##           .toc           list of toc paths for this subsets
##           .usr.blksize   space needed in /usr
##           .var.blksize   space needed in /var
##
##	product_load_details{} creates similar multi-field entries for
##	each product it loads.  The indices for these entries have the
##	form "<prod_name>.<field>" where prod_name is a product name
##	such as "Digital_UNIX", and field is one of:
##
##           Product Field  Description
##           -------------  ---------------------------------------------
##           .subset_list   list of names of all subsets in this product
##
##	Finally, there are some Subsets() entries for each file system,
##	summarizing status for all products.  The indices for these entries
##	have the form "<fs_name>.<field>" where fs_name is one of the usual
##	file system tags such as "usr" or "i18n", and field is one of:
##
##           F.S. Field  Description
##           ----------  ---------------------------------------------
##           .blktotal   total data size on this file system of all
##                       subsets currently selected for installation,
##                       in 512-byte blocks.
##           .fstotal    .blktotal plus estimated disk usage for other
##			 purposes (fitset RESERVE, non-inventory files
##			 such as /usr/.smdb./*, file system overhead, etc.
##           .sizestr    user-visible string representing .fstotal
##

global	CatNum
set CatNum 0  ;# when this file is sourced


## --------------------------------------------------------------------------
 # Proc: pcode_lower
 #
 # Inputs:
 #	ss_name - name of a subset
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	the first three letters of the subset name (the product code),
 #	converted to lowercase.
 #
 # Notes:
 #	This utility routine returns the "pcode" (product code -- the
 #	three letter prefix) of a subset name, converted to lowercase.
## --------------------------------------------------------------------------
proc pcode_lower {ss_name} {

	return [string tolower [csubstr $ss_name 0 3]]

} ; ## End pcode_lower

## --------------------------------------------------------------------------
 # Proc: product_load_details
 #
 # Inputs:
 #	product - name of product for which we need details.
 #
 # Outputs:
 #	Updates the global arrays Subsets() and Toc().
 #
 # Returns:
 #	A null string for success; otherwise an error message.
 #
 # Notes:
 #	Runs "setld -s" and loads the results in the Subsets() array.
 #	Updates the Toc() array, assumed to already exist, with the
 #	subset categories returned by "setld -s".
## --------------------------------------------------------------------------
proc product_load_details {product} {

	global	errorInfo \
		Handles \
		MandatoryCache \
		ROOT_DIR \
		SetldCtx \
		Subsets \
		Toc

	# We get the subset details for a product by running "setld -s".
	# This is an expensive operation.  If we've already done it once
	# for this product, in the current setld context, just recycle
	# those results.  What is the setld context?  It's the state of
	# any and all environment variables that affect the evaluation
	# of FLAGS expressions in subset .ctrl files.  All such variables
	# must be set correctly before running "setld -s", and their
	# collective state must be encoded in the SetldCtx() string.
	# This global is modified only by setld_already_done{}.
	#
	set err_msg ""
	if {[catch {set done_flag [setld_already_done \
		$product err_msg]} ]} {
		return $errorInfo
	}
	if {! [lempty $err_msg]} {
		return $err_msg
	}
	if {$done_flag} {
		subset_copy_mand_from_cache $product
		return ""
	}

	# Otherwise, we haven't yet run "setld -s" for this product, with
	# the current setld context.  Do so now.
	#
	set err_ret [setld_s_start $product]
	if {! [lempty $err_ret]} {
		return $err_ret
	}

	# If we're running setld in the background, the following can
	# be done while waiting for it to complete.  Otherwise, the
	# following can run after it is done.
	#
	set prod_ss_in_toc [product_subsets_in_toc $product]

	# Create the disk usage variables, if they don't already exist.
	#
	if {! [info exists Subsets(root.blktotal)]} {
		subset_init_fs_usage
	}
	catch {update}

	# Wait for setld to finish, then read the setld-s.outdone file,
	# storing its information in the MandatoryCache(), Subsets(),
	# and Toc() global arrays.
	#
	set err [setld_s_wait]
	catch {update}
	if {! [lempty $err]} {
		return $err
	}

	subset_copy_mand_from_cache $product
	catch {update}

	return ""  ;# success

} ; ## End product_load_details

## --------------------------------------------------------------------------
 # Proc: product_subsets_in_toc
 #
 # Inputs:
 #	product - name of product for which we need details.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of the subsets in the named product, according to
 #	the distribution's table of contents (TOC) file.
 #
 # Notes:
## --------------------------------------------------------------------------
proc product_subsets_in_toc {product} {

	global	Toc

	# During a ris install, the client may not be registered for
	# all products in the TOC file.
	#
	if {! [info exists Toc($product)]} {
		return {}
	}

	set ss_list {}
	foreach prod_path $Toc($product) {
		set ss_list [union $ss_list [toc_list_childless $prod_path]]
	}

	return $ss_list

} ; ## End product_subsets_in_toc

## --------------------------------------------------------------------------
 # Proc: setld_already_done
 #
 # Inputs:
 #	product - name of a software product.
 #	err_ret_name - name of variable in calling routine in which
 #		we'll return any error message.
 #
 # Outputs:
 #	SetldCtx - A string representing the current setld context
 #		is assigned to this global.  This proc is the only
 #		place where SetldCtx is modified.
 #
 # Returns:
 #	True (1) if "setld -s" has previously been run in this context;
 #	otherwise false (0).
 #
 # Notes:
 #	Sets the context string for the mandatories cache from the
 #	values given in the argument list.
## --------------------------------------------------------------------------
proc setld_already_done {product {err_ret_name {}}} {

	global	MandatoryCache \
		SetldCtx \
		Toc
	upvar 1 $err_ret_name err_ret

	set err_ret ""
	if {! [info exists Toc($product.pcode)]} {
		set SetldCtx($product) "undef"
		return 0
	}
	set pcode $Toc($product.pcode)

	# Store in prod_func the name of the function that creates
	# the "setld -s" context string for this product.
	#
	set prod_func ${pcode}_setld_context

	# Call this product's context function.  Some products, such
	# as WLS do not define a context function.  We use "catch" to
	# avoid a stack dump in that case.  It would be cleaner to
	# first check for existence of the function using "info procs"
	# but that does the wrong thing for products that *do* have a
	# context function that hasn't yet been autoloaded.
	#
	set err_flag 0
	set SetldCtx(_error_) ""
	if {[catch {set context [$prod_func]} err]} {

		# Some kind of error occurred.  Immediately save the
		# stack dump info before anything overwrites it.
		#
		global	errorInfo
		set prod_func_err $errorInfo

		# Determine if the error is simply that this product
		# lacks a context function.  Because we've tried to
		# call it, we can be sure it's been autoloaded (if
		# it exists).  That means it's now correct to use
		# "info procs" to test its existence.
		#
		set product_func_list [info procs $prod_func]
		if {[lsearch -exact $product_func_list $prod_func] >= 0} {
			set err_flag 1
		} else {
			set context "_const_ctx_"
		}

	} elseif {! [lempty $SetldCtx(_error_)]} {

		# Report any error from the context function.
		#
		set err_flag 1
		set prod_func_err $SetldCtx(_error_)
	}

	if {$err_flag} {

		# Report the error from the context function.
		#
		set efmt [mcat_err_gets 400 \
			SUBSET_product_function_error COMMON]
		set err_ret [format $efmt $prod_func $prod_func_err]

		set context "_error_ctx_"
	}

	set SetldCtx($product) $context

	set pctx $product.$context
	set ret [info exists MandatoryCache($pctx.num_mand)]

	return $ret

} ; ## End setld_already_done

## --------------------------------------------------------------------------
 # Proc: setld_load_details
 #
 # Inputs:
 #	product - name of product we just ran "setld -s" for
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A null string on success; otherwise an error message.
 #
 # Notes:
## --------------------------------------------------------------------------
proc setld_load_details {product} {

	global	errorInfo

	if {[catch {set prod_ss_in_ctrl \
		[setld_load_outdone_file $product err_text]} ]} {
		return $errorInfo
	}
	if {! [lempty $err_text]} {
		return $err_text
	}

	subset_insert_toc_category $product $prod_ss_in_ctrl
	subset_hide_deprecated $product

	return ""  ;# success

} ; ## End setld_load_details

## --------------------------------------------------------------------------
 # Proc: setld_load_outdone_file
 #
 # Inputs:
 #	product - name of product for which we need details.
 #	err_ret_name - name of variable in calling routine in which
 #		we'll return any error message.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of the subset names found in the outdone file.
 #
 # Notes:
## --------------------------------------------------------------------------
proc setld_load_outdone_file {product err_ret_name} {

	global	FileName \
		Limits \
		LogFp \
		MandatoryCache \
		SetldCtx \
		Subsets
	upvar 1 $err_ret_name err_ret
	set err_ret ""

	# Load the "setld -s" output file.
	#
	set pctx $product.$SetldCtx($product)
	if {! [cequal $LogFp "stderr"]} {
		puts $LogFp "Loading $FileName(outdone) for $pctx"
	}
	set got_outdone 0
	for {set retry 1} {$retry <= 5} {incr retry} {

		# Attempt to read the entire outdone file into a
		# variable.  If the read succeeds, break out of
		# this retry loop.
		#
		if {! [catch {set outdone_data [read_file -nonewline \
			$FileName(outdone)]} err]} {
			set got_outdone 1
			break
		}

		# Report the failure in the log file, with some
		# additional diagnostic info.
		#
		puts $LogFp "Attempt #$retry failed:  $err"
		foreach dbg_cmd [list \
			"ls -CFl $FileName(outdone_dir)" \
			"df -k" \
		] {
			puts $LogFp "+ $dbg_cmd"
			if {[catch {set ss_dir [eval exec $dbg_cmd]} err]} {
				puts $LogFp "$err\n"
			} else {
				puts $LogFp "$ss_dir\n"
			}
		}
	}

	# If we failed to load the outdone file, complain to the user.
	#
	if {! $got_outdone} {
		set efmt [mcat_err_gets 400 SETLD_no_outdone_file_fmt COMMON]
		set err_ret [format $efmt $FileName(outdone) $err]
		return {}
	}

	# We were able to read the outdone file.
	#
	set outdone [split $outdone_data "\n"]
	if {[catch {set num_lines [llength $outdone]}]} {
		set num_lines "(unknown)"
	}
	if {! [cequal $LogFp "stderr"]} {
		puts $LogFp "Read $num_lines lines from $FileName(outdone)"
	}

	# If we had to retry the outdone file, report this to the user.
	# Hopefully this will generate a useful QAR.
	#
	if {$retry > 1} {
		efmt [mcat_err_gets 100 SETLD_retry_outdone_file_fmt COMMON]
		set warn_txt [format $efmt $FileName(outdone) \
			$FileName(log_file) $FileName(setld_log)]
		puts stderr $warn_txt
		if {! [cequal $LogFp "stderr"]} {
			puts $LogFp $warn_txt
		}
	}

	set MandatoryCache($pctx.num_mand) 0

	set line_num 0
	set num_errors 0
	set NUM_FIELDS 9
	set prod_ss_in_ctrl {}
	set counter 0
	foreach ssod $outdone {
		incr line_num
		if {$counter <= 0} {
			catch {update}
			set counter 50
		}
		incr counter -1

		# Abandon the file if a line isn't a valid tcl list.
		#
		if {[catch {set rec_len [llength $ssod]} ]} {
			set efmt [mcat_err_gets 400 \
				SETLD_outdone_garbled COMMON]
			append err_ret [format $efmt $line_num \
				$FileName(outdone) $ssod]
			incr num_errors
			break
		}

		# Reject this subset if record has wrong number of fields.
		#
		if {$rec_len != $NUM_FIELDS} {
			set efmt [mcat_err_gets 400 \
				SETLD_outdone_num_fields COMMON]
			append err_ret [format $efmt $NUM_FIELDS $rec_len \
				$line_num $FileName(outdone) $ssod]
			incr num_errors
			continue
		}

		# Parse the info for the current subset.
		#
		# Notice that the description string (desc) is ignored
		# in normal circumstances, because we'll use the one in
		# the TOC instead.  The only time we use this one is if
		# we have to create a new Toc() entry (i.e., if the TOC
		# data file omits a subset for which we have a .ctrl
		# file); see below.
		#
		lassign $ssod name \
			desc \
			catdesc \
			deps \
			root_bytes \
			usr_bytes \
			var_bytes \
			mand \
			status

		# Reject this subset if setld could not get all its data.
		#
		if {! [lempty $status]} {
			set efmt [mcat_err_gets 400 \
				SETLD_outdone_bad_status COMMON]
			append err_ret [format $efmt $status $line_num \
				$FileName(outdone) $ssod]
			incr num_errors
			continue
		}

		set Subsets($name.req_list) $deps

		# Record whether this subset is mandatory (given the
		# current setld context).
		#
		set MandatoryCache($pctx.$name) [cequal $mand "mandatory"]
		if {$MandatoryCache($pctx.$name)} {
			incr MandatoryCache($pctx.num_mand)
		}

		# If we've previously run "setld -s" on this product,
		# don't muck up that data by re-recording it.  All
		# that can have changed since the earlier run is the
		# mandatory status of the subsets, which we've just
		# made note of in the cache.
		#
		if {[info exists Subsets($product.subset_list)]} {
			continue
		}

		set bsize $Limits(disk_block_size)
		set i18n_bytes 0
		if {[string first "IOS" $name] == 0} {
			set i18n_bytes $usr_bytes
			set usr_bytes 0
		}
		set Subsets($name.i18n.blksize) [expr $i18n_bytes / $bsize]
		set Subsets($name.root.blksize) [expr $root_bytes / $bsize]
		set Subsets($name.usr.blksize) [expr $usr_bytes / $bsize]
		set Subsets($name.var.blksize) [expr $var_bytes / $bsize]

		# Set a few of the Subsets fields.
		#
		set Subsets($name.categ) $catdesc
		set Subsets($name.desc) $desc
		subset_define $name
		set Subsets($name.status) $status

		# Tell this subset's dependents that it depends upon them,
		# by adding it to each one's .enables list.
		#
		foreach dep $Subsets($name.req_list) {
			subset_define $dep
			set Subsets($dep.enables) \
				[union $Subsets($dep.enables) $name]
		}

		# Add this subset to the list of subsets provided by this
		# product.  The list may include duplicates (if a subset
		# appears in more than one place in the TOC); we'll get
		# rid of duplicates down below.
		#
		lappend prod_ss_in_ctrl $name

	} ; ## End:  foreach ssod $outdone

	if {$num_errors > 0} {
		catch {unset MandatoryCache($pctx.num_mand)}
		set prod_ss_in_ctrl {}
	}

	set prod_ss_in_ctrl [lrmdups $prod_ss_in_ctrl]

	# Create the list of subsets in this product, if we don't have
	# it already from a previous call to this routine.
	#
	if {! [info exists Subsets($product.subset_list)]} {

		# Save the list of subsets for this product.
		#
		set Subsets($product.subset_list) $prod_ss_in_ctrl
	}

	return $prod_ss_in_ctrl

} ; ## End setld_load_outdone_file

## --------------------------------------------------------------------------
 # Proc: setld_s_start
 #
 # Inputs:
 #	product - name of product for which we need details.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	An empty string for success; otherwise an error message.
 #
 # Notes:
## --------------------------------------------------------------------------
proc setld_s_start {product} {

	global	BASE_PROD_NAME \
		env \
		FileName \
		Handles \
		ROOT_DIR \
		SetldCtx

	# If we already have a "setld -s" process running, check if
	# it's for this product and context.  If so, return success.
	# Otherwise, wait for it to complete before starting a new one.
	#
	set start_pctx $product.$SetldCtx($product)
	set prior_pctx $SetldCtx(_active_)
	if {! [lempty $prior_pctx]} {

		if {[cequal $prior_pctx $start_pctx]} {
			return ""
		}

		# Wait for the prior "setld -s" to finish, and save its
		# results before starting a new one.
		#
		set wait_err [setld_s_wait]
		if {! [lempty $wait_err]} {
			return $wait_err
		}
	}

	set env(ADVFLAG) "1"

	set env(SETLD_PROD_NAME) $product
	set prod_ctrl_dir "$ROOT_DIR/$product/kit"

	# Backward-compatibility ugliness:  accept "BASE" as the
	# name of the base O.S. product, and translate as necessary
	# to make it work.  This translation can be removed when setld
	# is updated to understand the new base O.S. name.
	#
	if {[lsearch [list BASE $BASE_PROD_NAME] $product] >= 0} {
		set env(SETLD_PROD_NAME) BASE
		set prod_ctrl_dir $FileName(base_prod_dir)
	}

	# We're about to run "setld -s".  If this is a CD installation,
	# we need to pass the kit directory as argument.
	#
	set srcloc [ItmMgr $Handles(Inst_islinfo) get srcloc]
	if {! [regexp {:$} $srcloc]} {
		#
		# CD install.
		#
		set srcloc $prod_ctrl_dir
	}

	set SetldCtx(_active_) $start_pctx
	catch {file delete -force $FileName(outdone)}

	if {[info exists env(SETLD_S_CACHE)]} {
		set SetldCtx(dbg.pctx) $start_pctx
		foreach ss_archive {/outdone /tmp} {
			set cached $ss_archive/$start_pctx
			if {! [file exists $cached]} {
				continue
			}
			set ss_dir [file dirname $FileName(outdone)]
			if {! [file isdirectory $ss_dir]} {
				mkdir -path $ss_dir
			}
			if {! [catch {file copy -force $cached \
				$FileName(outdone)} cp_err]} {
				return ""
			}
		}
	}

	system "setld -s $srcloc &"

	return ""  ;# success

} ; ## End setld_s_start

## --------------------------------------------------------------------------
 # Proc: setld_s_wait
 #
 # Inputs:
 #	FileName(outdone) - name of output file created by "setld -s"
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A null string on success; otherwise an error message.
 #
 # Notes:
 #	Does not return until the outdone file exists.  Will return
 #	at once if the file exists when this routine is called.
## --------------------------------------------------------------------------
proc setld_s_wait {} {

	global	env \
		FileName \
		SetldCtx

	wait_for_process_done "setld" $FileName(outdone) \
		"/file/does/not/exist" err

	if {[info exists env(SETLD_S_CACHE)]} {
		set start_pctx $SetldCtx(dbg.pctx)
		set ss_archive /tmp
		set save_name $ss_archive/$start_pctx
		if {! [file exists $save_name]} {
			file copy -force $FileName(outdone) $save_name
		}
	}

	# Read the setld.outdone file, storing its information in the
	# MandatoryCache(), Subsets(), and Toc() global arrays.
	#
	if {! [lempty $SetldCtx(_active_)]} {
		lassign [split $SetldCtx(_active_) "."] product
		set err [setld_load_details $product]
		set SetldCtx(_active_) ""
	}

	return $err

} ; ## End setld_s_wait

## --------------------------------------------------------------------------
 # Proc: subset_copy_mand_from_cache
 #
 # Inputs:
 #	product - name of a software product.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
 #	Sets the mandatory status -- Subsets($name.mand) boolean -- of
 #	every subset in the named product to the state indicated by the
 #	mandatory cache for this product, and the current setld context.
 #	Assumes the cache has previously been initialized by a call to
 #	product_load_details{}.
## --------------------------------------------------------------------------
proc subset_copy_mand_from_cache {product} {

	global	MandatoryCache \
		SetldCtx \
		Subsets

	# If there are no subsets in this product, do nothing.
	#
	if {! [info exists Subsets($product.subset_list)]} {
		return
	}
	if {[lempty $Subsets($product.subset_list)]} {
		return
	}

	# Copy into the Subsets() array the mandatory/optional
	# status of every subset in this product.
	#
	set pctx $product.$SetldCtx($product)
	foreach name $Subsets($product.subset_list) {

		# If we encounter a subset name for which there is no
		# entry in the current cache context, create one that
		# says the subset is optional.  This situation should
		# never arise, assuming the TOC file lists the same
		# subsets as "setld -s".
		#
		if {! [info exists MandatoryCache($pctx.$name)]} {
			set MandatoryCache($pctx.$name) 0  ;# never happen?
		}

		set Subsets($name.mand) $MandatoryCache($pctx.$name)
	}

	catch [pcode_lower $name]_load_details

} ; ## End subset_copy_mand_from_cache

## --------------------------------------------------------------------------
 # Proc: subset_define
 #
 # Inputs:
 #	name - name of subset to be defined.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
 #	Initializes the .mand and .sflag fields for the named subset.
 #	Creates a trace on .sflag to keep the disk usage numbers up to
 #	date.
## --------------------------------------------------------------------------
proc subset_define {name} {

	global	Subsets

	if {! [info exists Subsets($name.mand)]} {
		set Subsets($name.mand) 0
	}

	if {! [info exists Subsets($name.sflag)]} {
		set Subsets($name.sflag) 0
	}

	if {! [info exists Subsets($name.enables)]} {
		set Subsets($name.enables) {}
	}

	set existing_traces [trace vinfo Subsets($name.sflag)]
	if {! [regexp subset_pick_handler $existing_traces]} {
		trace variable Subsets($name.sflag) w subset_pick_handler
	}

} ; ## End subset_define

## --------------------------------------------------------------------------
 # Proc: subset_fast_flags
 #
 # Inputs:
 #	product - name of product.
 #	context - current setld context string for the product.
 #	shell_init - shell process initialization commands (may be empty).
 #	subset_list - names of subsets whose FLAGS need to be evaluated.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.  Updates the MandatoryCache() global array.
 #
 # Notes:
 #	This routine is called to avoid unnecessary use of the slow
 #	"setld -s" API for determining which subsets are mandatory.
 #	A subset is mandatory if the FLAGS expression in its .ctrl
 #	file evaluates to an integer in which the second bit is zero
 #	(typically the integer is 0 or 4).  An optional subset has
 #	a one in the second bit of its FLAGS value (typically the
 #	integer is 2 or 6).
 #
 #	The FLAGS expression in each .ctrl file is written in bourne
 #	shell notation.  This routine creates a shell process and
 #	writes commands on its standard input to evaluate the FLAGS
 #	for each subset.  It then reads the shell's standard output to
 #	get the results back, and stores them in the MandatoryCache()
 #	global array.
## --------------------------------------------------------------------------
proc subset_fast_flags {product context shell_init subset_list} {

	global	FileName \
		MandatoryCache \
		ROOT_DIR

	# Create the shell process, connected by a pipe.  Because we
	# open the pipe with mode "a+", the fd file handle will access
	# both stdin and stdout for the shell process.
	#
	if {[catch {set fd [open "|sh" a+]} err]} {
		set efmt [mcat_err_gets 400 SUBSET_cannot_pipe_to_setld COMMON]
		puts stderr [format $efmt $err]
		return
	}

	# Imitate the "setld -s" environment.
	#
	foreach sh_init_cmd {
		"DEFPATH=1"
		"CMDSW=-s"
		"ADVFLAG=1"
		"UPDFLAG="
	} {
		puts $fd $sh_init_cmd
	}
	set instctrl_dir $ROOT_DIR/$product/kit/instctrl
	puts $fd "L_DIR=$instctrl_dir"
	puts $fd "export DEFPATH CMDSW ADVFLAG UPDFLAG L_DIR"

	# Send any product-specific shell initialization commands.
	#
	puts $fd $shell_init

	# Get the FLAGS expression from the .ctrl file of each of
	# subset.  This statement sets the variable flags_expressions
	# to a list of shell command lines, each of the form:
	#    echo "subset_name=flags expression"
	#
	set start_dir [pwd]
	cd $instctrl_dir
	set ctrl_files [join $subset_list ".ctrl "].ctrl
	if {[catch {set flags_expressions [eval exec \
		grep {{^[[:space:]]*FLAGS=}} $ctrl_files | \
		sed \
		-e {{s!^[^:]*/!!}} \
		-e {{s!^\([^:]*\)\.ctrl!\1!}} \
		-e {{s!^\([^:]*\):[[:space:]]*FLAGS=\(.*\)!\1 echo \1=\2!}} \
		-e {{/$L_SUB/s!^\([^ ]*\)\( .*\)$L_SUB!\1\2\1!}} \
		-e {{s![^ ]* !!}} \
	]} err]} {
		set efmt [mcat_err_gets 300 SUBSET_no_flags COMMON]
		puts stderr [format $efmt [pwd] $err]
		continue
	}

	# Pass all the FLAGS expressions to the shell process for
	# execution.  The shell will produce output lines of the form:
	#   subset_name=value
	# where value is a small integer.  If the second bit of the
	# value integer is set, the subset is optional.
	#
	puts $fd $flags_expressions

	cd $start_dir

	puts $fd "exit 0"
	flush $fd

	set pctx $product.$context
	while {[gets $fd result] >= 0} {

		# Get the subset name and FLAGS value from this
		# line of shell output.
		#
		lassign [split $result "="] subset flags
		if {[lempty $flags]} {
			set flags 0
		}

		# Extract bit two of the FLAGS value; if it is
		# a one, this subset is optional.
		#
		set is_opt [expr ($flags / 2) % 2]

		# If this subset's status in the current context
		# differs from the preceding context that we used
		# as our rough cut beginning, update the mandatory
		# count to allow for the true status.
		#
		set is_mand [expr 1 - $is_opt]
		if {$is_mand != $MandatoryCache($pctx.$subset)} {

			set MandatoryCache($pctx.$subset) $is_mand

			if {$is_mand} {
				incr MandatoryCache($pctx.num_mand)
			} else {
				incr MandatoryCache($pctx.num_mand) -1
			}
		}

		# Special check for end of file -- we don't want a
		# "gets" to hang just because the pipe remains open.
		#
		if {[eof $fd]} {
			break
		}
	}

	# Close the pipe to kill the shell process.
	#
	if {[catch {close $fd} err]} {
		set efmt [mcat_err_gets 300 SUBSET_pipe_close_error COMMON]
		puts stderr [format $efmt $err]
	}

} ; ## End subset_fast_flags

## --------------------------------------------------------------------------
 # Proc: subset_hide_deprecated
 #
 # Inputs:
 #	product - name of a software product.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
 #	Delete from the Toc() any subsets for which "setld -s" reported
 #	no info.  These are typically subsets where the menu phase of
 #	the .scp returned failure, meaning: don't display this subset
 #	to the user.  This is done by, e.g., the kernel object module
 #	subsets, which are very large and needed only by kernel developers.
 #	Taking them out of the Toc() ensures that even if the user picks
 #	"All" base subsets, these large and little-used subsets will not
 #	be installed.  Users who *really* want them will have to run
 #	setld manually, after installation.
## --------------------------------------------------------------------------
proc subset_hide_deprecated {product} {

	global	Subsets \
		Toc

	foreach subset [product_subsets_in_toc $product] {

		# If this subset was reported by "setld -s", skip it.
		#
		if {[info exists Subsets($subset.req_list)]} {
			continue
		}

		# Otherwise, it must be a subset for which there's a
		# TOC entry, but it isn't reported by "setld -s".  That
		# means the menu phase of its .scp is saying this subset
		# should not be shown to the user.  So here we delete it
		# from the Toc(), so that we won't show it.
		#
		foreach toc_path [toc_find $subset] {
			toc_destroy $toc_path
		}
	}

} ; ## End subset_hide_deprecated

## --------------------------------------------------------------------------
 # Proc: subset_init_fs_usage
 #
 # Inputs:
 #	None
 #
 # Outputs:
 #	Initializes the Subsets($fs.blktotal) variables.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_init_fs_usage {} {

	global	FStags \
		Subsets

	# Create all the variables.  We create them all before setting
	# any traces, to ensure that all are available for use by the
	# trace handler, no matter which is modified first.
	#
	foreach fs $FStags(fs) {
		set Subsets($fs.blktotal) 0
	}

} ; ## End subset_init_fs_usage

## --------------------------------------------------------------------------
 # Proc: subset_insert_toc_category
 #
 # Inputs:
 #	product - name of a software product.
 #	subset_list - list of subset names
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_insert_toc_category {product subset_list} {

	global	CatNum \
		Subsets \
		Toc

	foreach name $subset_list {

		# Determine all TOC paths to this subset.  If we don't find
		# any, it means this subset is not in the TOC.  In that
		# case we add it to the Toc() array, as a member of the
		# current product, not in any package.  This may or may
		# not be correct, but it is reasonable and it makes the
		# subset available for installation.  For the description
		# string, we use the "desc" value from the .ctrl file.
		#
		set toc_path_list [toc_find $name]
		if {[lempty $toc_path_list]} {
			set efmt [mcat_err_gets 200 TOC_entry_not_found COMMON]
			puts stderr [format $efmt $name]
			if {! [info exists Toc($product)] ||
				! [info exists Subsets($name.desc)]} {
				continue
			}
			lassign $Toc($product) prod_path
			toc_create $prod_path $name $Subsets($name.desc) {}
			set toc_path_list $prod_path/$name
		}

		set catdesc $Subsets($name.categ)
		foreach toc_path $toc_path_list {

			regsub "/$name" $toc_path "" parent
			set tc_ix "$parent/$catdesc"
			if {[info exists tmpcats($tc_ix)]} {
				set catname $tmpcats($tc_ix)
			} else {
				set catname CATEGORY_$CatNum
				incr CatNum
				set tmpcats($tc_ix) $catname

				toc_create $parent $catname $catdesc {}
			}

			# Move this subset into its category.
			#
			toc_move $toc_path $parent/$catname
			set toc_path $parent/$catname/$name

			# Record the TOC path in the Subsets() array.  We
			# use lappend rather than set because subsets can
			# appear more than once in the TOC.  For example
			# IOSWWPRINT500 is listed in numerous WLS packages.
			#
			lappend Subsets($name.toc) $toc_path

		} ; ## End:  foreach toc_path $toc_path_list

	} ; ## End:  foreach name $subset_list

} ; ## End subset_insert_toc_category

## --------------------------------------------------------------------------
 # Proc: subset_list_all_dependencies
 #
 # Inputs:
 #	ss_list - a list of subset names
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of all subsets that are dependencies, at any remove,
 #	of any of the ss_list subsets.
 #
 # Notes:
 #	Recursively traces the complete dependency tree for each
 #	subset in the list.
## --------------------------------------------------------------------------
proc subset_list_all_dependencies {ss_list} {

	return [subset_list_all_related $ss_list req_list]

} ; ## End subset_list_all_dependencies

## --------------------------------------------------------------------------
 # Proc: subset_list_all_enabled
 #
 # Inputs:
 #	ss_list - a list of subset names
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of all subsets, at any remove, that have one or more
 #	of the ss_list subsets as a dependency.
 #
 # Notes:
 #	Recursively traces the complete dependency tree for each
 #	subset in the list.
## --------------------------------------------------------------------------
proc subset_list_all_enabled {ss_list} {

	return [subset_list_all_related $ss_list enables]

} ; ## End subset_list_all_enabled

## --------------------------------------------------------------------------
 # Proc: subset_list_all_related
 #
 # Inputs:
 #	ss_list - a list of subset names
 #	ss_field - either "req_list" or "enables"
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of all subsets that are dependencies, at any remove,
 #	of any of the ss_list subsets.
 #
 # Notes:
 #	Recursively traces the complete dependency tree for each
 #	subset in the list.
## --------------------------------------------------------------------------
proc subset_list_all_related {ss_list ss_field} {

	global	Subsets

	set dep_list {}
	foreach subset $ss_list {

		if {! [info exists Subsets($subset.$ss_field)]} {
			continue
		}

		foreach depstr $Subsets($subset.$ss_field) {

			# See if depstr contains any regular expression
			# metacharacters.
			#
			if {! [regexp {[[*?]} $depstr]} {

				# depstr is just an ordinary string.
				# Add it to the list.
				#
				lappend dep_list $depstr
				continue
			}

			# depstr contains metacharacters -- it must
			# be a regular expression.  Match it against
			# the universe of known subsets.
			#
			set dep_match [array names Subsets "$depstr.$ss_field"]
			regsub -all ".$ss_field" $dep_match {} dep_ss
			set dep_list [concat $dep_list $dep_ss]
		}
	}
	set dep_list [lrmdups $dep_list]

	lassign [intersect3 $dep_list $ss_list] only_new
	if {! [lempty $only_new]} {

		# Recursively check for additional related subsets.
		#
		set deps_of_new [subset_list_all_related $only_new $ss_field]

		set dep_list [union $dep_list $deps_of_new]
	}

	return $dep_list

} ; ## End subset_list_all_related

## --------------------------------------------------------------------------
 # Proc: subset_list_auto_related
 #
 # Inputs:
 #	ss_list - a list of subset names
 #	pick_flag - true if the subsets are being selected for installation;
 #		false if they're being de-selected.
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	A list of all other subsets that should get the same pick operation.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_list_auto_related {ss_list pick_flag} {

	global	Subsets

	if {$pick_flag} {

		# Make cfai ("Check For Auto-Inclusion") a list of all
		# subsets that are prerequisites for any subset in
		# ss_list.  Some or all of the cfai subsets may also
		# be in ss_list.
		#
		set cfai [subset_list_all_dependencies $ss_list]

	} else {

		# Make cfai a list of all subsets that depend upon any
		# subset in ss_list.  Some or all of the cfai subsets
		# may also be in ss_list.
		#
		set cfai [subset_list_all_enabled $ss_list]
	}

	# Make can_auto a list of only subsets from cfai that
	# are not already in ss_list.
	#
	lassign [intersect3 $cfai $ss_list] can_auto

	# Make must_auto a list of all subsets in the can_auto list
	# whose pick state differs from pick_flag.  Some or all of
	# these can_auto subsets may also be in ss_list.
	#
	set must_auto {}
	foreach name $can_auto {
		if {($pick_flag && ! $Subsets($name.sflag)) ||
			(! $pick_flag && $Subsets($name.sflag))} {
			lappend must_auto $name
		}
	}

	return $must_auto

} ; ## End subset_list_auto_related

## --------------------------------------------------------------------------
 # Proc: subset_pick_handler
 #
 # Inputs:
 #	This is a standard tcl "trace" handler:
 #	   always_Subsets - names the traced global array
 #	   name_plus - a subset name with ".sflag" appended
 #	   always_w - the trace is triggered only for writes
 #
 # Outputs:
 #	None.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_pick_handler {always_Subsets name_plus always_w} {

	global	Subsets

	if {! [regsub {\.sflag$} $name_plus {} ss_name]} {
		return
	}

	set amt -1
	if {$Subsets($name_plus)} {
		set amt 1
	}

	foreach toc_path [toc_find $ss_name] {  ;# toc_find{} here is *slow*
		toc_incr $toc_path pkd_ss $amt
	}

} ; ## End subset_pick_handler

## --------------------------------------------------------------------------
 # Proc: subset_pick_list
 #
 # Inputs:
 #	pick_flag - true if the subsets are being selected for installation;
 #		false if they're being de-selected.
 #	ss_list - names of subsets affected by the operation.
 #
 # Outputs:
 #	Updates the global Subsets() array .sflag field of each affected
 #	subsets.  Also updates the file system total usage values.
 #
 # Returns:
 #	A list of the subsets that were automatically included in the
 #	operation.  These will be dependencies of ss_list if pick_flag
 #	is true; otherwise they will be subsets that depend upon the
 #	ss_list subsets.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_pick_list {pick_flag ss_list} {

	global	FStags \
		Subsets

	# Make direct a list of the ss_list entries whose .sflag does
	# not already match the pick_flag.
	#
	set direct {}
	foreach name $ss_list {
		if {! [info exists Subsets($name.sflag)]} {
			subset_define $name
		}

		# We don't simply check for pick_flag!=.sflag here,
		# because .sflag may become a bitmask (where each of
		# the bits would be a flag representing a reason for
		# the subset to be selected).
		#
		if {($pick_flag && ! $Subsets($name.sflag)) ||
			(! $pick_flag && $Subsets($name.sflag))} {
			lappend direct $name
		}
	}

	# If every subset in ss_list currently has the desired pick
	# state, then we have nothing to do.
	#
	if {[lempty $direct]} {
		return {}
	}

	# Otherwise, direct is a list of subsets that need to change
	# pick state (to match pick_flag).  Determine what additional
	# subsets must be included in this operation.
	#
	set must_auto [subset_list_auto_related $direct $pick_flag]

	# Add the must_auto subsets to the list of subsets given
	# initially for this operation.
	#
	set work_list [union $direct $must_auto]

	# Mark each subset as indicated by the pick_flag.  Total
	# the resulting change in disk usage on each file system.
	#
	foreach fs $FStags(fs) {
		set du_blocks($fs) 0
	}
	foreach name $work_list {

		set Subsets($name.sflag) $pick_flag

		if {[info exists Subsets($name.root.blksize)]} {
			foreach fs $FStags(fs) {
				incr du_blocks($fs) $Subsets($name.$fs.blksize)
			}
		}
	}

	# Create the /tmp/.smdb. directory and create/remove temporary 
	# lock files for use by the WLS subsets to determine if they
	# should be mandatory or optional.
	#
	set ios_smdb /tmp/.smdb.
	catch {mkdir -path $ios_smdb}
	foreach ss $work_list {
		set path $ios_smdb/$ss.lk
		if {$pick_flag} {
			catch {close [open $path w]}
		} else {
			file delete -force -- $path
		}
	}

	# Update the overall file system usage totals.
	#
	set sign "-"
	if {$pick_flag} {
		set sign ""
	}
	foreach fs $FStags(fs) {
		incr Subsets($fs.blktotal) $sign$du_blocks($fs)
		subset_update_fstotal $fs
	}

	# Return the list of automatically included subsets, if any.
	#
	return $must_auto

} ; ## End subset_pick_list

## --------------------------------------------------------------------------
 # Proc: subset_update_all_fstotals
 #
 # Inputs:
 #	None.
 #
 # Outputs:
 #	Updates the global Subsets() array .fstotal field for every
 #	file system.
 #
 # Returns:
 #	Nothing.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_update_all_fstotals {} {

	global	FStags

	foreach fs $FStags(fs) {
		subset_update_fstotal $fs
	}

} ; ## End subset_update_all_fstotals

## --------------------------------------------------------------------------
 # Proc: subset_update_fstotal
 #
 # Inputs:
 #	fs - file system tag, e.g. "usr"
 #
 # Outputs:
 #	Updates the global Subsets() array .fstotal field for the named
 #	file system.
 #
 # Returns:
 #	The new Subsets($fs.fstotal) value in 512-blocks.
 #
 # Notes:
## --------------------------------------------------------------------------
proc subset_update_fstotal {fs} {

	global	FSreserve \
		Limits \
		Subsets

	# Get the name of the file system, and its .blktotal
	#
	set ix "$fs.blktotal"
	if {! [info exists Subsets($ix)]} {
		return
	}
	set blktotal $Subsets($ix)

	# If this file system is in /usr, update the /usr size info.
	#
	lassign [get_fs_info $fs] fsdev
	set merge_fs 0

	if {$merge_fs && [cequal $fsdev "in_usr"]} {

		# Because this file system is "in /usr", its own
		# .fstotal must be zero.
		#
		set Subsets($fs.fstotal) 0
		set Subsets($fs.sizestr) [block_count_to_str \
			$Subsets($fs.fstotal)]

		# Make blktotal the sum of the .blktotal's for /usr
		# and any other file systems that are "in /usr".
		#
		set fs usr
		set blktotal $Subsets(usr.blktotal)
		
		foreach opt_fs {var i18n} {
			lassign [get_fs_info $opt_fs] opt_fsdev
			if {! [cequal $opt_fsdev "in_usr"]} {
				continue
			}
			incr blktotal $Subsets($opt_fs.blktotal)
			incr blktotal [expr $FSreserve($opt_fs) \
				* $Limits(blocks_per_kb)]
		}
	}

	# Update our estimated disk usage for this file system,
	# based on its .blktotal -- the total subset data size
	# assigned to it.
	#
	if {[catch {set Subsets($fs.fstotal) [estimate_size_in_fs \
		$fs $blktotal]} err]} {
		global	errorInfo
		lassign [info level [info level]] myname
		set efmt [mcat_err_gets 300 SUBSET_estimation_error COMMON]
		puts stderr [format $efmt $myname $errorInfo]
		set Subsets($fs.fstotal) $Subsets($fs.blktotal)
	}

	# Create a string representing the .fstotal size in a format
	# suitable for display to the user.
	#
	set Subsets($fs.sizestr) [block_count_to_str $Subsets($fs.fstotal)]

} ; ## End subset_update_fstotal
